#include "Map.h"

#include "wProblem.h"
#include "wMshData.h"
#include "wElement.h"
#include "wNode.h"
#include "wGroup.h"
#include "wTag.h"

#include <mpi.h>

#include "wContact.h"

#include <Teuchos_Array.hpp>

using namespace katoptron;
using namespace tbox;

void std_vector_intersection(std::vector<Map::global_ordinal_type> vec1,
                             std::vector<int> vec2, std::vector<int> &vec3)
{
    std::sort(vec2.begin(), vec2.end());
    for (auto i = 0; i < vec1.size(); ++i)
    {
        if (std::find(vec2.begin(), vec2.end(), vec1[i]) != vec2.end())
            vec3.push_back(vec1[i]);
    }
}

/**
 * @brief Map constructor
 * 
 * This constructor initializes the maps used to describe the distributed matrices and vectors.
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: A Problem object,
 *  <li> numPrimalDPN: the number of degrees of freedom per node (without taking into account the Lagrange multipliers).
 * </ul>
 */
Map::Map(Problem &pbl, size_t _numPrimalDPN)
{
    using Teuchos::Array;
    using Teuchos::ArrayRCP;
    using Teuchos::ArrayView;
    using Teuchos::RCP;
    using Teuchos::rcp;

    numPrimalDPN = _numPrimalDPN;

    RCP<const Teuchos::Comm<int>> comm = rcp(new Teuchos::MpiComm<int>(pbl.comm));

    const size_t myRank = comm->getRank();
    const size_t numtasks = comm->getSize();

    auto msh = pbl.msh;

    const Tpetra::global_size_t numNodes = msh->nodes.size();
    const Tpetra::global_size_t numPrimalDOFs = numNodes * numPrimalDPN;
    const Tpetra::global_size_t numThermalDOFs = numNodes;
    const Tpetra::global_size_t numMechanicalDOFs = numNodes * 3;

    // ------------------------------------------------------------- //
    //                     Node and elements maps
    // ------------------------------------------------------------- //

    // Nodes map with overlaps
    // Create a map which associates to each mpi process the nodes included
    // in their elements partition. This map has overlaps associated to the
    // interfaces of the partitions.
    std::vector<global_ordinal_type> myNodesWO1 = {};
    std::vector<global_ordinal_type> myNodesWO2 = {};
    std::vector<global_ordinal_type> tmpMyNodesWO1 = {};
    std::vector<global_ordinal_type> tmpMyNodesWO2 = {};
    std::vector<global_ordinal_type> myElems = {};

    if (numtasks == 1)
    {
        for (auto e : msh->elems)
            if (e->type() != ElType::POINT1)
                myElems.push_back(e->no - 1);

        for (auto n : msh->nodes)
        {
            myNodesWO1.push_back(n->row);
            myNodesWO2.push_back(n->row);
        }
    }
    else
    {
        for (auto const &p : msh->parts)
        {
            if (p.second->no == (myRank + 1))
            {
                for (auto e : p.second->elems)
                {
                    if (e->type() == ElType::HEX8 || e->type() == ElType::TETRA4)
                        for (auto n1 : e->nodes)
                        {
                            tmpMyNodesWO1.push_back(n1->row);
                            tmpMyNodesWO2.push_back(n1->row);
                        }
                    if (e->type() != ElType::POINT1)
                        myElems.push_back(e->no - 1);
                }
            }
            else if (p.second->no == -(myRank + 1))
            {
                for (auto e : p.second->elems)
                {
                    if (e->type() == ElType::HEX8 || e->type() == ElType::TETRA4)
                        for (auto n1 : e->nodes)
                            tmpMyNodesWO1.push_back(n1->row);

                    if (e->type() != ElType::POINT1)
                        myElems.push_back(e->no - 1);
                }
            }
        }

        std::sort(tmpMyNodesWO1.begin(), tmpMyNodesWO1.end());
        std::sort(tmpMyNodesWO2.begin(), tmpMyNodesWO2.end());
        std::sort(myElems.begin(), myElems.end());

        myNodesWO1.push_back(tmpMyNodesWO1[0]);

        for (auto i = 1; i < tmpMyNodesWO1.size(); ++i)
            if (tmpMyNodesWO1[i] != tmpMyNodesWO1[i - 1])
                myNodesWO1.push_back(tmpMyNodesWO1[i]);

        myNodesWO2.push_back(tmpMyNodesWO2[0]);

        for (auto i = 1; i < tmpMyNodesWO2.size(); ++i)
            if (tmpMyNodesWO2[i] != tmpMyNodesWO2[i - 1])
                myNodesWO2.push_back(tmpMyNodesWO2[i]);
    }

    mapNodesWO = rcp(new tpetra_map_type(numNodes, myNodesWO1, indexBase, comm));
    mapElems = rcp(new tpetra_map_type(Teuchos::OrdinalTraits<Tpetra::global_size_t>::invalid(), myElems, indexBase, comm));

    // Loop on the contact and remove from myNodesWO2 any nodes included in a contact boundary for which the
    // current mpi process has no contact element.
    // This prevent Moertel of not finding the nodes.
    int contacts_number = pbl.Contacts.size();
    if (numtasks != 1)
        for (auto i = 0; i < contacts_number; ++i)
        {
            {
                std::vector<int> v3;
                std_vector_intersection(myNodesWO2, pbl.Contacts[i]->slave_boundary_nodes, v3);
                for (auto j = 0; j < pbl.Contacts[i]->slave_elems.size(); ++j)
                    for (auto jj = 0; jj < pbl.Contacts[i]->slave_elems[j]->nodes.size(); ++jj)
                    {
                        auto no = pbl.Contacts[i]->slave_elems[j]->nodes[jj]->row;

                        auto p = std::find(v3.begin(), v3.end(), no);
                        if (p != v3.end())
                        {
                            v3.erase(p);
                        }
                    }

                for (auto j = 0; j < v3.size(); ++j)
                {
                    auto p = std::find(myNodesWO2.begin(), myNodesWO2.end(), v3[j]);
                    if (p != myNodesWO2.end())
                    {
                        myNodesWO2.erase(p);
                    }
                }
            }
            {
                std::vector<int> v3;
                std_vector_intersection(myNodesWO2, pbl.Contacts[i]->master_boundary_nodes, v3);

                for (auto j = 0; j < pbl.Contacts[i]->master_elems.size(); ++j)
                    for (auto jj = 0; jj < pbl.Contacts[i]->master_elems[j]->nodes.size(); ++jj)
                    {
                        auto no = pbl.Contacts[i]->master_elems[j]->nodes[jj]->row;
                        auto p = std::find(v3.begin(), v3.end(), no);
                        if (p != v3.end())
                        {
                            v3.erase(p);
                        }
                    }
                for (auto j = 0; j < v3.size(); ++j)
                {
                    auto p = std::find(myNodesWO2.begin(), myNodesWO2.end(), v3[j]);
                    if (p != myNodesWO2.end())
                    {
                        myNodesWO2.erase(p);
                    }
                }
            }
        }
    RCP<const Tpetra::Map<>> mapNodesWO_contact = rcp(new tpetra_map_type(numNodes, myNodesWO2, indexBase, comm));

    // Create a map of nodes without overlaps. Each node is owned by one and
    // only mpi process.
    mapNodes = Tpetra::createOneToOne(mapNodesWO_contact);

    // Create a map of the degrees of fredom associated to the nodes of the
    // previous map.
    const size_t numMyNodes = mapNodes->getNodeNumElements();
    const size_t numMyNodesWO = mapNodesWO->getNodeNumElements();
    Teuchos::ArrayView<const global_ordinal_type> myNodes = mapNodes->getNodeElementList();

    // ------------------------------------------------------------- //
    //                          Dofs maps
    // ------------------------------------------------------------- //

    // Create a map of the degrees of fredom associated to the nodes of the
    // previous map.
    Array<global_ordinal_type> myDofsWO(numMyNodesWO * numPrimalDPN);
    Array<global_ordinal_type> myThermalDofsWO(numMyNodesWO);
    Array<global_ordinal_type> myMechanicalDofsWO(numMyNodesWO * 3);

    for (auto i = 0; i < numMyNodesWO; ++i)
    {
        if (numPrimalDPN == 4)
        {
            myThermalDofsWO[i] = myNodesWO1[i];
            myDofsWO[i] = myNodesWO1[i];

            for (auto j = 0; j < 3; ++j)
            {
                myDofsWO[numMyNodesWO + 3 * i + j] = 3 * myNodesWO1[i] + j + numNodes;
                myMechanicalDofsWO[3 * i + j] = 3 * myNodesWO1[i] + j;
            }
        }
        else if (numPrimalDPN == 3)
        {
            for (auto j = 0; j < 3; ++j)
            {
                myMechanicalDofsWO[3 * i + j] = 3 * myNodesWO1[i] + j;
                myDofsWO[3 * i + j] = 3 * myNodesWO1[i] + j;
            }
        }
        else if (numPrimalDPN == 1)
        {
            myThermalDofsWO[i] = myNodesWO1[i];
            myDofsWO[i] = myNodesWO1[i];
        }
    }

    mapDofsWO = rcp(new tpetra_map_type(numPrimalDOFs, myDofsWO, indexBase, comm));
    if (numPrimalDPN == 4 || numPrimalDPN == 1)
        thermoMapWO = rcp(new tpetra_map_type(numThermalDOFs, myThermalDofsWO, indexBase, comm));
    if (numPrimalDPN == 4 || numPrimalDPN == 3)
        mechaMapWO = rcp(new tpetra_map_type(numMechanicalDOFs, myMechanicalDofsWO, indexBase, comm));

    Array<global_ordinal_type> myDofs(numMyNodes * numPrimalDPN);
    Array<global_ordinal_type> myThermalDofs(numMyNodes);
    Array<global_ordinal_type> myMechanicalDofs(numMyNodes * 3);
    Array<global_ordinal_type> myThermomechanicalDofs(numMyNodes * numPrimalDPN);

    for (auto i = 0; i < numMyNodes; ++i)
    {
        if (numPrimalDPN == 4)
        {
            myThermalDofs[i] = myNodes[i];
            myDofs[i] = myNodes[i];

            for (auto j = 0; j < 3; ++j)
            {
                myMechanicalDofs[3 * i + j] = 3 * myNodes[i] + j;
                myDofs[numMyNodes + 3 * i + j] = 3 * myNodes[i] + j + numNodes;
            }
        }
        else if (numPrimalDPN == 3)
        {
            for (auto j = 0; j < 3; ++j)
            {
                myMechanicalDofs[3 * i + j] = 3 * myNodes[i] + j;
                myDofs[3 * i + j] = 3 * myNodes[i] + j;
            }
        }
        else if (numPrimalDPN == 1)
        {
            myThermalDofs[i] = myNodes[i];
            myDofs[i] = myNodes[i];
        }
    }

    mapDofs = rcp(new tpetra_map_type(numPrimalDOFs, myDofs, indexBase, comm));
    if (numPrimalDPN == 4 || numPrimalDPN == 1)
        thermoMap = rcp(new tpetra_map_type(numThermalDOFs, myThermalDofs, indexBase, comm));
    if (numPrimalDPN == 4 || numPrimalDPN == 3)
        mechaMap = rcp(new tpetra_map_type(numMechanicalDOFs, myMechanicalDofs, indexBase, comm));

    if (numPrimalDPN == 4)
    {
        thermoMap = rcp(new tpetra_map_type(numThermalDOFs, myThermalDofs, indexBase, comm));
        mechaMap = rcp(new tpetra_map_type(numMechanicalDOFs, myMechanicalDofs, indexBase, comm));

        thermomecFullmap = rcp(new tpetra_map_type(numPrimalDOFs, myThermomechanicalDofs, indexBase, comm));
        RCP<xpetra_map_type> xthermomecFullmap = rcp(new xpetra_tmap_type(thermomecFullmap));
        std::vector<RCP<const xpetra_map_type>> xsubmaps = {rcp(new xpetra_tmap_type(thermoMap)),
                                                            rcp(new xpetra_tmap_type(mechaMap))};

        thermomecBlockedMap = rcp(new xpetra_bmap_type(xthermomecFullmap, xsubmaps, true));
    }
    else if (numPrimalDPN == 3)
    {
        mechaMap = rcp(new tpetra_map_type(numMechanicalDOFs, myMechanicalDofs, indexBase, comm));

        thermomecFullmap = rcp(new tpetra_map_type(numPrimalDOFs, myThermomechanicalDofs, indexBase, comm));
        RCP<xpetra_map_type> xthermomecFullmap = rcp(new xpetra_tmap_type(thermomecFullmap));
        std::vector<RCP<const xpetra_map_type>> xsubmaps = {rcp(new xpetra_tmap_type(mechaMap))};

        thermomecBlockedMap = rcp(new xpetra_bmap_type(xthermomecFullmap, xsubmaps, true));
    }
    else
    {
        thermoMap = rcp(new tpetra_map_type(numThermalDOFs, myThermalDofs, indexBase, comm));

        thermomecFullmap = rcp(new tpetra_map_type(numPrimalDOFs, myThermomechanicalDofs, indexBase, comm));
        RCP<xpetra_map_type> xthermomecFullmap = rcp(new xpetra_tmap_type(thermomecFullmap));
        std::vector<RCP<const xpetra_map_type>> xsubmaps = {rcp(new xpetra_tmap_type(thermoMap))};

        thermomecBlockedMap = rcp(new xpetra_bmap_type(xthermomecFullmap, xsubmaps, true));
    }

    // Construct a Map that puts all the equations on processor 0 (for outputs).
    const size_t numLclIndices = (myRank == 0) ? numPrimalDOFs : 0;

    mapOutput = rcp(new tpetra_map_type(numPrimalDOFs, numLclIndices, indexBase, comm));
}