#include "wContact.h"
#include "wProblem.h"
#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
#include <algorithm>

#include <fstream>
#include <sstream>
#include <string>

using namespace katoptron;
using namespace tbox;

bool searchCode(char const *code, char const *line)
{
    return !(strncmp(code, line, strlen(code)));
}

bool searchCode(char const *code, std::string const line)
{
    return !(strncmp(code, line.c_str(), strlen(code)));
}

bool searchCode(std::string const code, std::string const line)
{
    return line.find(code) != std::string::npos;
}

void readGMSH(std::string const &file_name, std::string const &name, std::vector<int> &nodes)
{
    std::cout << file_name << ": start " << std::endl;
    std::vector<int> nodes_tmp;
    std::ifstream infile(file_name);
    int tag = 0;
    std::string line;
    bool reading_physical_names = false;
    while (std::getline(infile, line))
    {
        if (reading_physical_names)
        {
            if (searchCode("$EndPhysicalNames", line))
                break;

            if (searchCode(name, line))
            {
                std::istringstream line_stream(line);
                int dim = 0, no = 0;
                line_stream >> dim >> no;
                tag = no;
                std::cout << "tag: " << tag << std::endl;
                reading_physical_names = false;
            }
        }
        else if (searchCode("$Elements", line))
        {
            std::getline(infile, line);
            std::istringstream line_stream(line);
            // read number of elements
            int nbelm = 0;

            line_stream >> nbelm;

            for (int i = 0; i < nbelm; ++i)
            {
                std::getline(infile, line);
                std::istringstream line_stream(line);

                // read number, type and nb of tags
                int no = 0, type = 0, nbtags = 0;
                int tmp;

                line_stream >> no >> type >> nbtags;
                // read/create/link tags

                bool do_I_continue = true;
                for (int j = 0; j < nbtags; ++j)
                {
                    line_stream >> tmp;
                    if (tag == tmp)
                        do_I_continue = false;
                }
                if (do_I_continue)
                    continue;

                int n = 0;
                // read nodes
                if (type == static_cast<int>(ElType::HEX8)) // 8-node hexa
                    n = 8;
                else if (type == static_cast<int>(ElType::QUAD4)) // 4-node quad
                    n = 4;
                else if (type == static_cast<int>(ElType::TETRA4)) // 4-node tetra
                    n = 4;
                else if (type == static_cast<int>(ElType::LINE2)) // 2-node line
                    n = 2;
                else if (type == static_cast<int>(ElType::POINT1)) // 1-node point
                    n = 1;
                else if (type == static_cast<int>(ElType::TRI3)) // 3-node triangle
                    n = 3;
                else
                {
                    continue;
                }
                int nn;
                for (int j = 0; j < n; ++j)
                {
                    line_stream >> nn;
                    nodes_tmp.push_back(nn - 1);
                }
            }
        }
        else if (searchCode("$EndElements", line))
        {
            std::cout << "nodes_tmp.size(): " << nodes_tmp.size() << std::endl;
            if (nodes_tmp.size() > 0)
            {
                std::sort(nodes_tmp.begin(), nodes_tmp.end());

                nodes.push_back(nodes_tmp[0]);

                for (auto i = 1; i < nodes_tmp.size(); ++i)
                    if (nodes_tmp[i] != nodes_tmp[i - 1])
                        nodes.push_back(nodes_tmp[i]);
            }
            std::cout << "nodes.size(): " << nodes.size() << std::endl;
            break;
        }
        else if (searchCode("$PhysicalNames", line))
        {
            reading_physical_names = true;
        }
    }
}

/**
 * @brief Contact constructor
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> no: the ID of the slave surface of the msh,
 *  <li> _Contact_name: the name of new contact interface (does not have to be a nametag of the msh),
 *  <li> _slave_normal: the slave normal,
 *  <li> dg: the initial displacement along the slave normal.
 * </ul>
 */
Contact::Contact(katoptron::Problem &pbl, int no, std::string _Contact_name, Eigen::Vector3d _slave_normal, double _dg) : Group(pbl.msh, no), Contact_name(_Contact_name), slave_normal(_slave_normal), dg(_dg)
{
    dg_vector = {dg};
    for (auto e : tag->elems)
        slave_elems.push_back(e);
}

/**
 * @brief Contact constructor
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> name: the name of the slave surface of the msh,
 *  <li> _Contact_name: the name of new contact interface (does not have to be a nametag of the msh),
 *  <li> _slave_normal: the slave normal,
 *  <li> dg: the initial displacement along the slave normal.
 * </ul>
 */
Contact::Contact(katoptron::Problem &pbl, std::string const &name, std::string _Contact_name, Eigen::Vector3d _slave_normal, double _dg) : Group(pbl.msh, name), Contact_name(_Contact_name), slave_normal(_slave_normal), dg(_dg)
{
    dg_vector = {dg};
    for (auto e : tag->elems)
        slave_elems.push_back(e);
}

/**
 * @brief Contact constructor
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> no: the ID of the slave surface of the msh,
 *  <li> _Contact_name: the name of new contact interface (does not have to be a nametag of the msh),
 *  <li> _slave_normal: the slave normal,
 *  <li> _dg: the initial displacement along the slave normal (value can be different for the different samples).
 * </ul>
 */
Contact::Contact(katoptron::Problem &pbl, int no, std::string _Contact_name, Eigen::Vector3d _slave_normal, std::vector<double> _dg) : Group(pbl.msh, no), Contact_name(_Contact_name), slave_normal(_slave_normal)
{
    dg = _dg[0];
    dg_vector = _dg;

    for (auto e : tag->elems)
        slave_elems.push_back(e);
}

/**
 * @brief Contact constructor
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> name: the name of the slave surface of the msh,
 *  <li> _Contact_name: the name of new contact interface (does not have to be a nametag of the msh),
 *  <li> _slave_normal: the slave normal,
 *  <li> _dg: the initial displacement along the slave normal (value can be different for the different samples).
 * </ul>
 */
Contact::Contact(katoptron::Problem &pbl, std::string const &name, std::string _Contact_name, Eigen::Vector3d _slave_normal, std::vector<double> _dg) : Group(pbl.msh, name), Contact_name(_Contact_name), slave_normal(_slave_normal)
{
    dg = _dg[0];
    dg_vector = _dg;

    for (auto e : tag->elems)
        slave_elems.push_back(e);
}

/**
 * @brief Set the master surface
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> no: the ID of the master surface of the msh,
 *  <li> _master_normal: the master normal.
 * </ul>
 */
void Contact::setMaster(katoptron::Problem &pbl, int no, Eigen::Vector3d _master_normal)
{
    master_normal = _master_normal;
    Group master_group(pbl.msh, no);
    Tag *master_tag = master_group.tag;
    for (auto e : master_tag->elems)
        master_elems.push_back(e);

    pbl.Contacts.push_back(this);
}

/**
 * @brief Set the master surface
 * 
 * Arguments:
 * <ul>
 *  <li> pbl: a Problem object,
 *  <li> name: the name of the master surface of the msh,
 *  <li> _master_normal: the master normal.
 * </ul>
 */
void Contact::setMaster(katoptron::Problem &pbl, std::string const &name, Eigen::Vector3d _master_normal)
{
    master_normal = _master_normal;
    Group master_group(pbl.msh, name);
    Tag *master_tag = master_group.tag;
    for (auto e : master_tag->elems)
        master_elems.push_back(e);

    pbl.Contacts.push_back(this);
}

/**
 * @brief Set slave boundary (optional)
 * 
 * Arguments:
 * <ul>
 *  <li> file_name: name of the msh file to use,
 *  <li> name: the name of the slave boundary of the msh.
 * </ul>
 */
void Contact::setSlaveBoundary(std::string const &file_name, std::string const &name)
{
    readGMSH(file_name, name, slave_boundary_nodes);

    std::cout << "slave_boundary_nodes.size(): " << slave_boundary_nodes.size() << std::endl;
}

/**
 * @brief Set master boundary (optional)
 * 
 * Arguments:
 * <ul>
 *  <li> file_name: name of the msh file to use,
 *  <li> name: the name of the master boundary of the msh.
 * </ul>
 */
void Contact::setMasterBoundary(std::string const &file_name, std::string const &name)
{
    readGMSH(file_name, name, master_boundary_nodes);

    std::cout << "master_boundary_nodes.size(): " << master_boundary_nodes.size() << std::endl;
}

/**
 * @brief Set slave Dirichlet boundary (optional)
 * 
 * Arguments:
 * <ul>
 *  <li> file_name: name of the msh file to use,
 *  <li> name: the name of the slave Dirichlet boundary of the msh.
 * </ul>
 */
void Contact::setSlaveDirichlet(std::string const &file_name, std::string const &name)
{
    readGMSH(file_name, name, slave_dirichlet_nodes);

    std::cout << "slave_dirichlet_nodes.size(): " << slave_dirichlet_nodes.size() << std::endl;
    for (auto i = 0; i < slave_dirichlet_nodes.size(); ++i)
        std::cout << slave_dirichlet_nodes[i] << std::endl;
}

/**
 * @brief Set master Dirichlet boundary (optional)
 * 
 * Arguments:
 * <ul>
 *  <li> file_name: name of the msh file to use,
 *  <li> name: the name of the master Dirichlet boundary of the msh.
 * </ul>
 */
void Contact::setMasterDirichlet(std::string const &file_name, std::string const &name)
{
    readGMSH(file_name, name, master_dirichlet_nodes);

    std::cout << "master_dirichlet_nodes.size(): " << master_dirichlet_nodes.size() << std::endl;
}

/**
 * @brief Activate all the Lagrange multipliers associated to 
 * the current contact interface.
 */
void Contact::setInitialyClosed()
{
    initial_closed = true;
}

/**
 * @brief Inactivate all the Lagrange multipliers associated to 
 * the current contact interface.
 */
void Contact::setInitialyOpen()
{
    initial_closed = false;
}

/**
 * @brief Specify to the active set strategy that this interface
 * should not be updated. 
 */
void Contact::setNoUpdate()
{
    no_Update = true;
}

/**
 * @brief Specify to the active set strategy that this interface
 * should be updated if required.
 */
void Contact::setUpdate()
{
    no_Update = false;
}

/**
 * @brief Replace this contact interface by a mesh tying interface. 
 */
void Contact::setMeshTying()
{
    mesh_tying = true;
    sticking = true;
    no_Update = true;
}

/**
 * @brief Specify that this interface is sticking. 
 */
void Contact::setSticking()
{
    sticking = true;
}

/**
 * @brief Specify that this interface is not sticking (default). 
 */
void Contact::setNoSticking()
{
    sticking = false;
}

/**
 * @brief Specify that the discretization of the Lagrange multipliers 
 * use standard shape functions (default).
 */
void Contact::setStandardShapeFunctions()
{
    standard_shape_functions = true;
}

/**
 * @brief Specify that the discretization of the Lagrange multipliers 
 * use dual shape functions.
 */
void Contact::setDualShapeFunctions()
{
    standard_shape_functions = false;
}

/**
 * @brief Specify that this interface is a Signorini contact.
 */
void Contact::setSignoriniContact()
{
    is_Signorini = true;
}

/**
 * @brief Specify that this interface is a not Signorini contact (default).
 */
void Contact::setTwoBodyContact()
{
    is_Signorini = false;
}

/**
 * @brief Specify a list of slave nodes where the Lagrange multipliers should
 * be initially closed.
 */
void Contact::setInitiallyClosedNodes(std::vector<double> _initially_closed_nodes)
{
    for (auto i = 0; i < _initially_closed_nodes.size(); ++i)
        initially_closed_nodes.push_back(((int)_initially_closed_nodes[i]));
}

/**
 * @brief Specify the scale.
 */
void Contact::setScale(double a)
{
    scale = a;
}

/**
 * @brief Display information.
 */
void Contact::write(std::ostream &out) const
{
    out << "Contact interface named ''" << Contact_name << "'' on " << *tag << "\n";
}